import os
from typing import Dict, Optional

try:
    import tiledb.cloud
except ImportError:
    pass

CLOUD_MODELS = "ml_models"
FILETYPE_ML_MODEL = "ml_model"


def get_s3_prefix(namespace: Optional[str]) -> Optional[str]:
    """
    Get S3 path from the user profile or organization profile.
    :param namespace: User's namespace
    :return: S3 prefix
    """
    if namespace is None:
        return None
    profile = tiledb.cloud.client.user_profile()
    if namespace == profile.username:
        s3_path = profile.default_s3_path
    else:
        s3_path = tiledb.cloud.client.organization(namespace).default_s3_path
    return os.path.join(s3_path, CLOUD_MODELS) if s3_path is not None else None


def update_file_properties(uri: str, file_properties: Dict[str, str]) -> None:
    tiledb.cloud.array.update_file_properties(
        uri=uri,
        file_type=FILETYPE_ML_MODEL,
        file_properties=file_properties,
    )


def get_cloud_uri(uri: str, namespace: Optional[str]) -> str:
    s3_prefix = get_s3_prefix(namespace)
    if s3_prefix is None:
        raise ValueError(
            f"You must set the default s3 prefix path for ML models in "
            f"{namespace} profile settings on TileDB-Cloud"
        )

    return f"tiledb://{namespace}/{os.path.join(s3_prefix, uri)}"



import pickle
from collections import Counter
from datetime import timedelta

import numpy as np
import pytest

from tiledb.ml.readers._ranges import InclusiveRange, IntRange, WeightedRange


@pytest.mark.parametrize("values", [None, 42, 3.14])
def test_inclusive_range_factory_type_error(values):
    with pytest.raises(TypeError) as excinfo:
        InclusiveRange.factory(values)
    assert "Cannot create inclusive range" in str(excinfo.value)


class TestIntRange:
    values = range(10, 20)
    r = InclusiveRange.factory(values)

    def test_basic(self):
        assert self.r.min == 10
        assert self.r.max == 19
        assert self.r.weight == 10
        assert len(self.r) == 10

    @pytest.mark.parametrize(
        "values",
        [
            values,
            list(values),
            set(values),
            iter(values),
            reversed(values),
            Counter(values),
            np.array(values),
            range(19, 9, -1),
            np.arange(19, 9, -1),
        ],
    )
    def test_equal(self, values):
        assert_equal_ranges(self.r, InclusiveRange.factory(values), IntRange)

    @pytest.mark.parametrize(
        "values",
        [
            np.array(values, dtype=object),
            range(0, 10),
            range(10, 21),
            range(11, 20),
            range(10, 20, 2),
        ],
    )
    def test_not_equal(self, values):
        r = InclusiveRange.factory(values)
        assert self.r != r
        assert not self.r.equal_values(r)

    def test_indices(self):
        np.testing.assert_array_equal(
            self.r.indices(np.array([10, 16, 16, 11, 19, 11])),
            np.array([0, 6, 6, 1, 9, 1]),
        )

    @pytest.mark.parametrize(
        "values",
        [
            [10, 16, 16, 11, 20],
            [10, 16, 16, 11, 9],
            [10, 16, 16, 11, 15.5],
        ],
    )
    def test_indices_error(self, values):
        with pytest.raises(ValueError) as excinfo:
            self.r.indices(np.array(values))
        assert "Values not in the range" in str(excinfo.value)

    @pytest.mark.parametrize(
        "k,expected_bounds",
        [
            (1, [(10, 20)]),
            (2, [(10, 15), (15, 20)]),
            (3, [(10, 14), (14, 17), (17, 20)]),
            (4, [(10, 13), (13, 16), (16, 18), (18, 20)]),
            (5, [(10, 12), (12, 14), (14, 16), (16, 18), (18, 20)]),
            (6, [(10, 12), (12, 14), (14, 16), (16, 18), (18, 19), (19, 20)]),
            (7, [(10, 12), (12, 14), (14, 16)] + [(i, i + 1) for i in range(16, 20)]),
            (8, [(10, 12), (12, 14)] + [(i, i + 1) for i in range(14, 20)]),
            (9, [(10, 12)] + [(i, i + 1) for i in range(12, 20)]),
            (10, [(i, i + 1) for i in range(10, 20)]),
        ],
    )
    def test_partition_by_count(self, k, expected_bounds):
        ranges = list(self.r.partition_by_count(k))
        assert len(ranges) == k
        expected_ranges = [InclusiveRange.factory(range(*bs)) for bs in expected_bounds]
        assert ranges == expected_ranges

    def test_partition_by_count_error(self):
        for k in range(11, 20):
            with pytest.raises(ValueError) as excinfo:
                list(self.r.partition_by_count(k))
            assert "Cannot partition range" in str(excinfo.value)

    @pytest.mark.parametrize(
        "max_weight,expected_bounds",
        [
            (1, [(i, i + 1) for i in range(10, 20)]),
            (2, [(10, 12), (12, 14), (14, 16), (16, 18), (18, 20)]),
            (3, [(10, 13), (13, 16), (16, 19), (19, 20)]),
            (4, [(10, 14), (14, 18), (18, 20)]),
            (5, [(10, 15), (15, 20)]),
            (6, [(10, 16), (16, 20)]),
            (7, [(10, 17), (17, 20)]),
            (8, [(10, 18), (18, 20)]),
            (9, [(10, 19), (19, 20)]),
            (10, [(10, 20)]),
            (11, [(10, 20)]),
        ],
    )
    def test_partition_by_weight(self, max_weight, expected_bounds):
        ranges = list(self.r.partition_by_weight(max_weight))
        assert max(r.weight for r in ranges) <= max_weight
        expected_ranges = [InclusiveRange.factory(range(*bs)) for bs in expected_bounds]
        assert ranges == expected_ranges

    def test_pickle(self):
        assert pickle.loads(pickle.dumps(self.r)) == self.r


class TestWeightedRange:
    values = ("e", "f", "a", "d", "a", "c", "d", "a", "f", "c", "f", "f", "b", "d")
    r = InclusiveRange.factory(values)
    r2 = InclusiveRange.factory({v: timedelta(c) for v, c in Counter(values).items()})

    @pytest.mark.parametrize("r", [r, r2])
    def test_basic(self, r):
        assert r.min == "a"
        assert r.max == "f"
        assert len(r) == 6
        assert r.weight == 14 if r is self.r else timedelta(14)

    @pytest.mark.parametrize(
        "values",
        [
            values,
            list(values),
            iter(values),
            reversed(values),
            Counter(values),
            np.array(values),
            np.array(values, dtype=object),
        ],
    )
    def test_equal(self, values):
        assert_equal_ranges(self.r, InclusiveRange.factory(values), WeightedRange)

    def test_not_equal(self):
        assert self.r != InclusiveRange.factory(set(self.values))
        assert self.r != InclusiveRange.factory(range(len(set(self.values))))

    def test_equal_values(self):
        assert self.r.equal_values(InclusiveRange.factory(set(self.values)))

        r = InclusiveRange.factory([1, 2, 3, 3, 4, 5])
        assert r.equal_values(InclusiveRange.factory(range(1, 6)))
        assert not r.equal_values(InclusiveRange.factory(range(1, 7)))
        assert not r.equal_values(InclusiveRange.factory(range(2, 7)))

    def test_strided_range(self):
        assert_equal_ranges(
            InclusiveRange.factory(range(10, 20, 3)),
            InclusiveRange.factory([10, 13, 16, 19]),
            WeightedRange,
        )
        assert_equal_ranges(
            InclusiveRange.factory(range(20, 10, -3)),
            InclusiveRange.factory([11, 14, 17, 20]),
            WeightedRange,
        )

    def test_indices(self):
        np.testing.assert_array_equal(
            self.r.indices(np.array(["a", "e", "e", "d", "a", "f", "c"])),
            np.array([0, 4, 4, 3, 0, 5, 2]),
        )

    @pytest.mark.parametrize(
        "values",
        [
            ["a", "e", "e", "d", "_"],
            ["a", "e", "e", "d", "z"],
            ["a", "e", "e", "d", "aa"],
        ],
    )
    def test_indices_error(self, values):
        with pytest.raises(ValueError) as excinfo:
            self.r.indices(np.array(values))
        assert "Values not in the range" in str(excinfo.value)

    parametrize_by_count = pytest.mark.parametrize(
        "k,expected_mappings",
        [
            (1, [{"e": 1, "f": 4, "a": 3, "d": 3, "c": 2, "b": 1}]),
            (2, [{"b": 1, "c": 2, "a": 3}, {"e": 1, "f": 4, "d": 3}]),
            (3, [{"b": 1, "a": 3}, {"c": 2, "d": 3}, {"e": 1, "f": 4}]),
            (4, [{"a": 3}, {"b": 1, "c": 2}, {"d": 3, "e": 1}, {"f": 4}]),
            (5, [{"a": 3}, {"b": 1, "c": 2}, {"d": 3}, {"e": 1}, {"f": 4}]),
            (6, [{"a": 3}, {"b": 1}, {"c": 2}, {"d": 3}, {"e": 1}, {"f": 4}]),
        ],
    )

    @parametrize_by_count
    def test_partition_by_count(self, k, expected_mappings):
        ranges = list(self.r.partition_by_count(k))
        assert len(ranges) == k
        expected_ranges = list(map(InclusiveRange.factory, expected_mappings))
        assert ranges == expected_ranges

    @parametrize_by_count
    def test_partition_by_count2(self, k, expected_mappings):
        ranges = list(self.r2.partition_by_count(k))
        assert len(ranges) == k
        expected_ranges = [
            InclusiveRange.factory({v: timedelta(w) for v, w in mapping.items()})
            for mapping in expected_mappings
        ]
        assert ranges == expected_ranges

    @pytest.mark.parametrize("r", [r, r2])
    def test_partition_by_count_error(self, r):
        for k in range(7, 20):
            with pytest.raises(ValueError) as excinfo:
                list(r.partition_by_count(k))
            assert "Cannot partition range" in str(excinfo.value)

    parametrize_by_max_weight = pytest.mark.parametrize(
        "max_weight,expected_mappings",
        [
            (4, [{"b": 1, "a": 3}, {"c": 2}, {"d": 3, "e": 1}, {"f": 4}]),
            (5, [{"b": 1, "a": 3}, {"c": 2, "d": 3}, {"e": 1, "f": 4}]),
            (6, [{"b": 1, "a": 3, "c": 2}, {"d": 3, "e": 1}, {"f": 4}]),
            (7, [{"b": 1, "a": 3, "c": 2}, {"d": 3, "e": 1}, {"f": 4}]),
            (8, [{"b": 1, "a": 3, "c": 2}, {"d": 3, "e": 1, "f": 4}]),
            (9, [{"b": 1, "a": 3, "c": 2, "d": 3}, {"e": 1, "f": 4}]),
            (10, [{"b": 1, "a": 3, "c": 2, "d": 3, "e": 1}, {"f": 4}]),
            (11, [{"b": 1, "a": 3, "c": 2, "d": 3, "e": 1}, {"f": 4}]),
            (12, [{"b": 1, "a": 3, "c": 2, "d": 3, "e": 1}, {"f": 4}]),
            (13, [{"b": 1, "a": 3, "c": 2, "d": 3, "e": 1}, {"f": 4}]),
            (14, [{"b": 1, "a": 3, "c": 2, "d": 3, "e": 1, "f": 4}]),
            (15, [{"b": 1, "a": 3, "c": 2, "d": 3, "e": 1, "f": 4}]),
        ],
    )

    @parametrize_by_max_weight
    def test_partition_by_weight(self, max_weight, expected_mappings):
        ranges = list(self.r.partition_by_weight(max_weight))
        assert max(r.weight for r in ranges) <= max_weight
        expected_ranges = list(map(InclusiveRange.factory, expected_mappings))
        assert ranges == expected_ranges

    @parametrize_by_max_weight
    def test_partition_by_weight2(self, max_weight, expected_mappings):
        max_weight = timedelta(max_weight)
        ranges = list(self.r2.partition_by_weight(max_weight))
        assert max(r.weight for r in ranges) <= max_weight
        expected_ranges = [
            InclusiveRange.factory({v: timedelta(w) for v, w in mapping.items()})
            for mapping in expected_mappings
        ]
        assert ranges == expected_ranges

    @pytest.mark.parametrize(
        "r,max_weights", [(r, range(1, 4)), (r2, map(timedelta, range(1, 4)))]
    )
    def test_partition_by_weight_error(self, r, max_weights):
        for max_weight in max_weights:
            with pytest.raises(ValueError):
                list(r.partition_by_weight(max_weight))

    def test_pickle(self):
        assert pickle.loads(pickle.dumps(self.r)) == self.r
        assert pickle.loads(pickle.dumps(self.r2)) == self.r2


def assert_equal_ranges(r1, r2, cls):
    assert isinstance(r1, cls)
    assert isinstance(r2, cls)
    assert r1.min == r2.min
    assert r1.max == r2.max
    assert r1.weight == r2.weight
    assert len(r1) == len(r2)
    assert r1 == r2
    assert r1.equal_values(r2)



import multiprocessing
import os

import pytest


@pytest.fixture(autouse=True, scope="session")
def set_multiprocessing_start_method():
    if os.name == "posix":
        multiprocessing.set_start_method("forkserver")



from tiledb.ml.models._cloud_utils import (
    get_cloud_uri,
    get_s3_prefix,
    update_file_properties,
)


class TestCloudUtils:
    def test_get_s3_prefix(self, mocker):

        profile = mocker.patch(
            "tiledb.cloud.client.user_profile",
            return_value=mocker.Mock(username="foo", default_s3_path="bar"),
        )
        profile.default_s3_path = "bar"
        assert get_s3_prefix("foo") == "bar/ml_models"

        profile = mocker.patch(
            "tiledb.cloud.client.user_profile",
            return_value=mocker.Mock(username="foo", default_s3_path=None),
        )
        profile.default_s3_path = "bar"
        assert get_s3_prefix("foo") is None

        org = mocker.patch(
            "tiledb.cloud.client.organization",
            return_value=mocker.Mock(default_s3_path="orgbar"),
        )
        org.default_s3_path = "orgbar"
        assert get_s3_prefix("nofoo") == "orgbar/ml_models"

        org = mocker.patch(
            "tiledb.cloud.client.organization",
            return_value=mocker.Mock(
                default_s3_path=None,
            ),
        )
        org.default_s3_path = "orgbar"
        assert get_s3_prefix("nofoo") is None

    def test_get_cloud_uri(self, mocker):
        mocker.patch(
            "tiledb.ml.models._cloud_utils.get_s3_prefix", return_value="s3://"
        )

        assert "tiledb://test_namespace/s3://tiledb_array" == get_cloud_uri(
            uri="tiledb_array", namespace="test_namespace"
        )

    def test_update_file_properties(self, mocker):
        mock_tiledb_cloud_update_file_properties = mocker.patch(
            "tiledb.cloud.array.update_file_properties"
        )

        uri = "tiledb_array"
        file_properties = {}

        update_file_properties(uri=uri, file_properties=file_properties)

        mock_tiledb_cloud_update_file_properties.assert_called_once_with(
            uri=uri, file_type="ml_model", file_properties=file_properties
        )



import itertools as it
import string

import numpy as np
import pytest

import tiledb
from tiledb.ml.readers._tensor_schema import DenseTensorSchema, SparseTensorSchema
from tiledb.ml.readers.types import ArrayParams


@pytest.fixture(scope="module")
def dense_uri(tmp_path_factory):
    uri = str(tmp_path_factory.mktemp("dense"))
    schema = tiledb.ArraySchema(
        sparse=False,
        domain=tiledb.Domain(
            tiledb.Dim(name="d0", domain=(0, 9999), dtype=np.int32, tile=123),
            tiledb.Dim(name="d1", domain=(-2, 2), dtype=np.int32, tile=2),
            tiledb.Dim(name="d2", domain=(1, 2), dtype=np.int32, tile=1),
        ),
        attrs=[
            tiledb.Attr(name="af8", dtype=np.float64),
            tiledb.Attr(name="af4", dtype=np.float32),
            tiledb.Attr(name="au1", dtype=np.uint8),
        ],
    )
    tiledb.Array.create(uri, schema)
    with tiledb.open(uri, "w") as a:
        a[:] = {
            "af8": np.random.rand(*schema.shape),
            "af4": np.random.rand(*schema.shape).astype(np.float32),
            "au1": np.random.randint(128, size=schema.shape, dtype=np.uint8),
        }
    return uri


@pytest.fixture(scope="module")
def sparse_uri(tmp_path_factory, non_zero_per_row=3):
    uri = str(tmp_path_factory.mktemp("sparse"))
    domains = [
        (1, 1000),
        (-250, 249),
        (0, 1),
        (np.datetime64("2020-01-01"), np.datetime64("2022-01-01")),
        ("a", "z"),
    ]
    schema = tiledb.ArraySchema(
        sparse=True,
        allows_duplicates=True,
        domain=tiledb.Domain(
            tiledb.Dim(name="d0", domain=domains[0], dtype=np.uint32),
            tiledb.Dim(name="d1", domain=domains[1], dtype=np.int16),
            tiledb.Dim(name="d2", domain=domains[2], dtype=np.float64),
            tiledb.Dim(name="d3", domain=domains[3], dtype=np.dtype("datetime64[D]")),
            tiledb.Dim(name="d4", domain=domains[4], dtype="ascii"),
        ),
        attrs=[
            tiledb.Attr(name="af8", dtype=np.float64),
            tiledb.Attr(name="af4", dtype=np.float32),
            tiledb.Attr(name="au1", dtype=np.uint8),
        ],
    )
    tiledb.Array.create(uri, schema)
    with tiledb.open(uri, "w") as a:
        d0 = np.concatenate(
            [np.arange(domains[0][0], domains[0][1] + 1)] * non_zero_per_row
        )
        num_cells = len(d0)
        d1 = np.random.randint(domains[1][0], domains[1][1] + 1, num_cells)
        d2 = np.random.uniform(domains[2][0], domains[2][1], num_cells)
        d3_days = (domains[3][1] - domains[3][0]).astype(int)
        d3 = domains[3][0] + np.random.randint(0, d3_days, num_cells)
        d4 = np.random.choice(list(string.ascii_lowercase), num_cells)
        a[d0, d1, d2, d3, d4] = {
            "af8": np.random.rand(num_cells),
            "af4": np.random.rand(num_cells).astype(np.float32),
            "au1": np.random.randint(128, size=num_cells, dtype=np.uint8),
        }
    return uri


def parametrize_fields(*fields):
    return pytest.mark.parametrize(
        "fields",
        list(it.chain.from_iterable(it.combinations(fields, i) for i in range(4))),
    )


@pytest.mark.parametrize(
    "key_dim_index,memory_budget",
    [(0, 16_000), (0, 32_000), (0, 64_000), (1, 500_000), (1, 600_000), (1, 700_000)],
)
@parametrize_fields("d0", "d1", "d2", "af8", "af4", "au1")
def test_max_partition_weight_dense(dense_uri, fields, key_dim_index, memory_budget):
    config = {"py.max_incomplete_retries": 0, "sm.memory_budget": memory_budget}
    with tiledb.open(dense_uri, config=config) as a:
        params = ArrayParams(a, key_dim_index, fields)
        schema = DenseTensorSchema.from_array_params(params)
        max_weight = schema.max_partition_weight
        for key_range in schema.key_range.partition_by_weight(max_weight):
            # query succeeds without incomplete retries
            schema.query[key_range.min : key_range.max]

            if key_range.max < schema.key_range.max:
                # querying a larger slice should fail
                with pytest.raises(tiledb.TileDBError) as ex:
                    schema.query[key_range.min : key_range.max + 1]
                assert "py.max_incomplete_retries" in str(ex.value)


@pytest.mark.parametrize("memory_budget", [2048, 4096])
@pytest.mark.parametrize("key_dim_index", [0, 1, 3, 4])
@parametrize_fields("d0", "d1", "d2", "d3", "d4", "af8", "af4", "au1")
def test_max_partition_weight_sparse(sparse_uri, fields, key_dim_index, memory_budget):
    config = {
        "py.max_incomplete_retries": 0,
        "py.init_buffer_bytes": memory_budget,
    }
    with tiledb.open(sparse_uri, config=config) as a:
        key_dim = a.dim(key_dim_index)
        params = ArrayParams(a, key_dim_index, fields)
        schema = SparseTensorSchema.from_array_params(params)
        max_weight = schema.max_partition_weight
        for key_range in schema.key_range.partition_by_weight(max_weight):
            # query succeeds without incomplete retries
            schema.query[key_range.min : key_range.max]

            if key_range.max < schema.key_range.max:
                # querying a larger slice should fail. `key_range.max + 1` makes sense
                # only for integer key dim so test just integer key dims
                if np.issubdtype(key_dim.dtype, np.integer):
                    with pytest.raises(tiledb.TileDBError) as ex:
                        schema.query[key_range.min : key_range.max + 1]
                    assert "py.max_incomplete_retries" in str(ex.value)



from __future__ import annotations

from abc import ABC, abstractmethod
from collections import Counter
from dataclasses import dataclass
from math import ceil
from operator import itemgetter
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
)

import numpy as np
import sparse

import tiledb

from ._ranges import InclusiveRange
from .types import ArrayParams

Tensor = TypeVar("Tensor")


@dataclass(frozen=True)  # type: ignore
class TensorSchema(ABC):
    """
    A class to encapsulate the information needed for mapping a TileDB array to tensors.
    """

    _array: tiledb.Array
    _key_dim_index: int
    _fields: Sequence[str]
    _all_dims: Sequence[str]
    _ned: Sequence[Tuple[Any, Any]]
    _query_kwargs: Dict[str, Any]
    _transform: Optional[Callable[[Tensor], Tensor]]

    @classmethod
    def from_array_params(
        cls,
        array_params: ArrayParams,
        transform: Optional[Callable[[Tensor], Tensor]] = None,
    ) -> TensorSchema:
        kwargs = {"_" + k: v for k, v in array_params._tensor_schema_kwargs.items()}
        return cls(_transform=transform, **kwargs)

    @property
    def fields(self) -> Sequence[str]:
        """Names of attributes and dimensions to read."""
        return self._fields

    @property
    def field_dtypes(self) -> Sequence[np.dtype]:
        """Dtypes of attributes and dimensions to read."""
        return tuple(map(self._array.schema.attr_or_dim_dtype, self._fields))

    @property
    def shape(self) -> Sequence[int]:
        """Shape of the array, with the key dimension moved first.

        **Note**: For sparse arrays, the returned shape reflects the non-empty domain of
        the array, not the full array shape.

        :raises ValueError: If the array does not have integer domain.
        """
        shape = [len(self.key_range)]
        for start, stop in self._ned[1:]:
            if isinstance(start, int) and isinstance(stop, int):
                shape.append(stop - start + 1)
            else:
                raise ValueError("Shape not defined for non-integer domain")
        return shape

    @property
    def query(self) -> KeyDimQuery:
        """A sliceable object for querying the TileDB array along the key dimension"""
        return KeyDimQuery(self._array, self._key_dim_index, **self._query_kwargs)

    @property
    @abstractmethod
    def sparse(self) -> bool:
        """Whether the underlying TileDB array is sparse"""

    @property
    @abstractmethod
    def key_range(self) -> InclusiveRange[Any, int]:
        """Inclusive range of the key dimension.

        The values of the range are all the distinct values of the key dimension (keys).
        The weight of each key is:
        - for dense arrays: 1
        - for sparse arrays: The number of non-empty cells for this key
        """

    @property
    @abstractmethod
    def max_partition_weight(self) -> int:
        """
        Determine the maximum partition that can be read without incomplete retries.

        What constitutes weight of a partition depends on the array type:
        - For dense arrays, it is the number of unique keys (= number of "rows").
          It depends on the `sm.memory_budget` config parameter.
        - For sparse arrays, it is the number of non-empty cells.
          It depends on the `py.init_buffer_bytes` config parameter.
        """

    @abstractmethod
    def iter_tensors(
        self, key_ranges: Iterable[InclusiveRange[Any, int]]
    ) -> Union[Iterable[Tensor], Iterable[Sequence[Tensor]]]:
        """
        Generate batches of dense or sparse tensors.

        Each yielded batch is either:
        - a sequence of N tensors if N > 1, where `N == len(self.fields)`, or
        - a single tensor if N == 1.
        where each tensor has shape `(len(key_range), *self.shape[1:])`.

        :param key_ranges: Inclusive ranges along the key dimension.
        """


class DenseTensorSchema(TensorSchema):
    sparse = False

    @property
    def key_range(self) -> InclusiveRange[int, int]:
        key_dim_min, key_dim_max = self._ned[0]
        return InclusiveRange.factory(range(key_dim_min, key_dim_max + 1))

    def iter_tensors(
        self, key_ranges: Iterable[InclusiveRange[int, int]]
    ) -> Union[Iterable[np.ndarray], Iterable[Sequence[np.ndarray]]]:
        """
        Generate batches of Numpy arrays.

        If `key_dim_index > 0`, the generated arrays will ve reshaped so that the key_dim
        axes is first. For example, if the TileDB array `a` has shape (5, 12, 20) and
        `key_dim_index == 1`, then `a[:, 4:8, :]` returns arrays of shape (5, 4, 20) but
        this method yields arrays of shape (4, 5, 20).
        """
        query = self.query
        get_data = itemgetter(*self._fields)
        key_dim_index = self._key_dim_index
        for key_range in key_ranges:
            field_arrays = query[key_range.min : key_range.max]
            if key_dim_index > 0:
                # Move key_dim_index axes first
                for field, array in field_arrays.items():
                    field_arrays[field] = np.moveaxis(array, key_dim_index, 0)
            yield get_data(field_arrays)

    @property
    def max_partition_weight(self) -> int:
        memory_budget = int(self._array._ctx_().config()["sm.memory_budget"])

        # The memory budget should be large enough to read the cells of the largest field
        bytes_per_cell = max(dtype.itemsize for dtype in self.field_dtypes)

        # We want to be reading tiles following the tile extents along each dimension.
        # The number of cells for each such tile is the product of all tile extents.
        dim_tiles = [self._array.dim(dim).tile for dim in self._all_dims]
        cells_per_tile = np.prod(dim_tiles)

        # Each slice consists of `rows_per_slice` rows along the key dimension
        rows_per_slice = dim_tiles[0]

        # Reading a slice of `rows_per_slice` rows requires reading a number of tiles that
        # depends on the size and tile extent of each dimension after the first one.
        assert len(self.shape) == len(dim_tiles)
        tiles_per_slice = np.prod(
            [ceil(size / tile) for size, tile in zip(self.shape[1:], dim_tiles[1:])]
        )

        # Compute the size in bytes of each slice of `rows_per_slice` rows
        bytes_per_slice = bytes_per_cell * cells_per_tile * tiles_per_slice

        # Compute the number of slices that fit within the memory budget
        num_slices = memory_budget // bytes_per_slice

        # Compute the total number of rows to slice
        return max(1, int(rows_per_slice * num_slices))


class SparseTensorSchema(TensorSchema):
    sparse = True

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        self._query_kwargs["dims"] = self._all_dims
        key_counter: Counter[Any] = Counter()
        key_dim = self._all_dims[0]
        query = self._array.query(dims=(key_dim,), attrs=(), return_incomplete=True)
        for result in query.multi_index[:]:
            key_counter.update(result[key_dim])
        self._key_range = InclusiveRange.factory(key_counter)

    @property
    def key_range(self) -> InclusiveRange[Any, int]:
        return self._key_range

    def iter_tensors(
        self, key_ranges: Iterable[InclusiveRange[Any, int]]
    ) -> Union[Iterable[Tensor], Iterable[Sequence[Tensor]]]:
        shape = list(self.shape)
        query = self.query
        get_data = itemgetter(*self._fields)
        single_field = len(self._fields) == 1
        key_dim, *non_key_dims = self._all_dims
        non_key_dim_starts = tuple(map(itemgetter(0), self._ned[1:]))
        transform = self._transform or (lambda x: x)
        for key_range in key_ranges:
            # Set the shape of the key dimension equal to the current key range length
            shape[0] = len(key_range)
            field_arrays = query[key_range.min : key_range.max]
            data = get_data(field_arrays)

            # Convert coordinates from the original domain to zero-based
            # For the key (i.e. first) dimension get the indices of the keys
            coords = [key_range.indices(field_arrays.pop(key_dim))]
            # For every non-key dimension, subtract the minimum value of the dimension
            # TODO: update this for non-integer non-key dimensions
            coords.extend(
                field_arrays.pop(dim) - dim_start
                for dim, dim_start in zip(non_key_dims, non_key_dim_starts)
            )

            # yield either a single tensor or a sequence of tensors, one for each field
            if single_field:
                yield transform(sparse.COO(coords, data, shape))
            else:
                yield tuple(transform(sparse.COO(coords, d, shape)) for d in data)

    @property
    def max_partition_weight(self) -> int:
        try:
            memory_budget = int(self._array._ctx_().config()["py.init_buffer_bytes"])
        except KeyError:
            memory_budget = 10 * 1024**2

        # The memory budget should be large enough to read the cells of the largest field
        bytes_per_cell = max(
            self._array.schema.attr_or_dim_dtype(field).itemsize
            for field in self._query_kwargs["dims"] + self._query_kwargs["attrs"]
        )
        return max(1, memory_budget // int(bytes_per_cell))


class KeyDimQuery:
    def __init__(self, array: tiledb.Array, key_dim_index: int, **kwargs: Any):
        self._multi_index = array.query(**kwargs).multi_index
        self._leading_dim_slices = (slice(None),) * key_dim_index

    def __getitem__(self, key_dim_slice: slice) -> Dict[str, np.ndarray]:
        """Query the TileDB array by `dim_key=key_dim_slice`."""
        return cast(
            Dict[str, np.ndarray],
            self._multi_index[(*self._leading_dim_slices, key_dim_slice)],
        )



from __future__ import annotations

import itertools as it
from abc import ABC, abstractmethod
from collections import Counter, abc
from dataclasses import dataclass
from functools import singledispatch
from typing import Any, Generic, Iterable, Mapping, NoReturn, TypeVar, Union, cast

import numpy as np

__all__ = ["InclusiveRange"]

V = TypeVar("V")
W = TypeVar("W")
VDtype = TypeVar("VDtype", bound=np.generic)
WDtype = TypeVar("WDtype", bound=np.number)


class InclusiveRange(ABC, Generic[V, W]):
    """
    Base abstract class for finite ranges that include both ends.

    An InclusiveRange `R` is a sorted set of comparable elements of type `V`. Every member
    `m` of `R` satisfies the inequality `R.min <= m <= R.max`.

    As a set, an InclusiveRange does not contain duplicates. However, range members may
    be associated with numeric weights of type `W`. This way duplicate members can be
    represented as unique members with weight equal to the number of occurrences.
    """

    __slots__: Iterable[str] = ()

    @property
    @abstractmethod
    def min(self) -> V:
        """Lower bound of this range."""

    @property
    @abstractmethod
    def max(self) -> V:
        """Upper bound of this range."""

    @property
    @abstractmethod
    def weight(self) -> W:
        """Total weight of the members of this range."""

    @property
    @abstractmethod
    def values(self) -> np.ndarray[V, Any]:
        """Unique sorted values of this range."""

    @abstractmethod
    def indices(self, values: np.ndarray[V, Any]) -> np.ndarray[np.int_, Any]:
        """Get the (0-based) indices of the given values in this range.

        :raises ValueError: If any of the given values is not in this range.
        """

    @abstractmethod
    def __len__(self) -> int:
        """Number of unique members in this range."""

    @abstractmethod
    def __eq__(self, other: Any) -> bool:
        """Check if this range equals to another object.

        Two ranges are equal if they consist of the same values with the same weights.
        """

    @abstractmethod
    def equal_values(self, other: InclusiveRange[V, W]) -> bool:
        """Check if two ranges consist of the same values."""

    @abstractmethod
    def partition_by_count(self, k: int) -> Iterable[InclusiveRange[V, W]]:
        """Partition this range into `k` subranges of approximately equal weight."""

    @abstractmethod
    def partition_by_weight(self, max_weight: W) -> Iterable[InclusiveRange[V, W]]:
        """
        Partition this range into the minimum number of subranges so that each subrange
        weight is at most `max_weight`.
        """

    @singledispatch
    @staticmethod
    def factory(values: Any) -> InclusiveRange[Any, Any]:
        """Create an inclusive range of comparable values.

        :param values: Can be:
        - an integer `range`. The weight of each value is 1.
        - an iterable of (potentially non-unique) values. The weight of each unique value
          is its cardinality.
        - a mapping from (unique) values to weights.
        """

    @staticmethod
    @factory.register(range)
    def _from_range(values: range) -> Union[IntRange, WeightedRange[np.int_, np.int_]]:
        if values.step < 0:
            values = range(values.stop + 1, values.start + 1, -values.step)

        if values.step == 1:
            return IntRange(values.start, values.stop - 1)
        else:
            return WeightedRange(
                np.arange(values.start, values.stop, values.step),
                np.ones(len(values), dtype=int),
            )

    @staticmethod
    @factory.register(abc.Iterable)
    def _from_iterable(values: Iterable[V]) -> Union[IntRange, InclusiveRange[V, int]]:
        return InclusiveRange._from_mapping(Counter(values))

    @staticmethod
    @factory.register(abc.Mapping)
    def _from_mapping(mapping: Mapping[V, W]) -> Union[IntRange, InclusiveRange[V, W]]:
        unique_sorted, weights = zip(*sorted(mapping.items()))
        return InclusiveRange._from_ndarrays(np.array(unique_sorted), np.array(weights))

    @staticmethod
    @factory.register(np.ndarray)
    def _from_ndarray(
        values: np.ndarray[VDtype, Any],
    ) -> Union[IntRange, WeightedRange[VDtype, np.int_]]:
        unique_sorted, counts = np.unique(values, return_counts=True)
        return InclusiveRange._from_ndarrays(unique_sorted, counts)

    @staticmethod
    def _from_ndarrays(
        unique_sorted: np.ndarray[VDtype, Any], weights: np.ndarray[WDtype, Any]
    ) -> Union[IntRange, WeightedRange[VDtype, WDtype]]:
        if (
            np.issubsctype(unique_sorted, np.integer)
            and unique_sorted[-1] - unique_sorted[0] + 1 == len(unique_sorted)
            and np.all(weights == 1)
        ):
            return IntRange(unique_sorted[0].item(), unique_sorted[-1].item())
        else:
            return WeightedRange(unique_sorted, weights)

    @staticmethod
    @factory.register(object)
    def _from_object(values: object) -> NoReturn:
        raise TypeError(f"Cannot create inclusive range from {values!r}")

    def __getstate__(self) -> Mapping[str, Any]:
        return {slot: getattr(self, slot) for slot in self.__slots__}

    def __setstate__(self, state: Mapping[str, Any]) -> None:
        for slot, value in state.items():
            object.__setattr__(self, slot, value)


@dataclass(frozen=True)
class IntRange(InclusiveRange[int, int]):
    """
    An inclusive range of consecutive integers.
    """

    __slots__ = ("min", "max")
    min: int
    max: int

    def __post_init__(self) -> None:
        assert self.min <= self.max

    @property
    def weight(self) -> int:
        return len(self)

    @property
    def values(self) -> np.ndarray[int, Any]:
        return np.arange(self.min, self.max + 1)

    def indices(self, values: np.ndarray[int, Any]) -> np.ndarray[np.int_, Any]:
        indices = values - self.min
        if not (
            np.issubsctype(indices, np.integer)
            and np.all(indices >= 0)
            and np.all(indices < len(self))
        ):
            raise ValueError(f"Values not in the range {self}")
        return indices

    def __len__(self) -> int:
        return self.max - self.min + 1

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, IntRange):
            return NotImplemented
        return self.min == other.min and self.max == other.max

    def equal_values(self, other: InclusiveRange[V, W]) -> bool:
        return self == other

    def partition_by_count(self, k: int) -> Iterable[IntRange]:
        n = len(self)
        if not (1 <= k <= n):
            raise ValueError(
                f"Cannot partition range of {n} members into {k} partitions"
            )
        d, m = divmod(n, k)
        # the first m partitions have length d+1 and the rest have length d
        lengths = it.chain(it.repeat(d + 1, m), it.repeat(d, k - m))
        yield from self._partition_by_lengths(lengths)

    def partition_by_weight(self, max_weight: int) -> Iterable[IntRange]:
        d, m = divmod(len(self), max_weight)
        # all partitions have length max_weight, with the possible exception of the last
        # partition that has length m (if m > 0)
        lengths = it.chain(it.repeat(max_weight, d), it.repeat(m, m > 0))
        yield from self._partition_by_lengths(lengths)

    def _partition_by_lengths(self, lengths: Iterable[int]) -> Iterable[IntRange]:
        start = self.min
        for length in lengths:
            next_start = start + length
            yield IntRange(start, next_start - 1)
            start = next_start


@dataclass(frozen=True)
class WeightedRange(InclusiveRange[VDtype, WDtype]):
    """An inclusive range of weighted comparable values."""

    __slots__ = ("values", "weights")
    values: np.ndarray[VDtype, Any]
    weights: np.ndarray[WDtype, Any]

    def __post_init__(self) -> None:
        assert self.values.ndim == self.weights.ndim == 1
        assert len(self.values) == len(self.weights)

    @property
    def min(self) -> VDtype:
        return cast(VDtype, self.values[0])

    @property
    def max(self) -> VDtype:
        return cast(VDtype, self.values[-1])

    @property
    def weight(self) -> WDtype:
        return cast(WDtype, np.sum(self.weights))

    def __len__(self) -> int:
        return len(self.values)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, WeightedRange):
            return NotImplemented
        return bool(
            len(self) == len(other)
            and np.all(self.values == other.values)
            and np.all(self.weights == other.weights)
        )

    def equal_values(self, other: InclusiveRange[V, W]) -> bool:
        return bool(len(self) == len(other) and np.all(self.values == other.values))

    def indices(self, values: np.ndarray[VDtype, Any]) -> np.ndarray[np.int_, Any]:
        members = self.values
        indices = np.searchsorted(members, values)
        if not (np.max(indices) < len(members) and np.all(members[indices] == values)):
            raise ValueError(f"Values not in the range {self}")
        return indices

    def partition_by_count(self, k: int) -> Iterable[WeightedRange[VDtype, WDtype]]:
        n = len(self)
        if not (1 <= k <= n):
            raise ValueError(
                f"Cannot partition range of {n} members into {k} partitions"
            )

        values, weights = self.values, self.weights
        if k == n:
            # edge case: one partition per value
            for i in range(n):
                yield WeightedRange(values[i : i + 1], weights[i : i + 1])
            return

        total_weight = self.weight
        # target_weight: sum of
        # (a) weights of the partitions yielded so far and
        # (b) the average of the remaining total weight over the remaining partitions
        target_weight = total_weight / k
        acc_weights = np.cumsum(weights)
        start = 0
        for i in range(k - 1, 0, -1):
            # find the index of target_weight in the accumulated weights
            stop = int(np.searchsorted(acc_weights, target_weight))
            # check if it's after the start index; if not, move it right after start
            if stop <= start < n:
                stop = start + 1
            # otherwise the following inequality holds:
            # acc_weights[stop - 1] < target_weight <= acc_weights[stop]
            elif (
                acc_weights[stop] - target_weight
                < target_weight - acc_weights[stop - 1]
            ):
                # increment stop if acc_weights[stop] is closer to target_weight than
                # acc_weights[stop - 1]
                stop += 1

            # yield partition
            yield WeightedRange(values[start:stop], weights[start:stop])

            # update target_weight and start index
            target_weight = acc_weights[stop - 1]
            target_weight += (total_weight - target_weight) / i
            start = stop

        # yield last partition
        yield WeightedRange(values[start:], weights[start:])

    def partition_by_weight(
        self, max_weight: WDtype
    ) -> Iterable[WeightedRange[VDtype, WDtype]]:
        values, weights = self.values, self.weights
        if max_weight < np.max(weights):
            raise ValueError(
                f"Cannot partition range with max weight={max_weight}: "
                f"max value weight={np.max(weights)}"
            )
        # target_weight: sum of weights of the partitions yielded so far plus max_weight
        target_weight = max_weight
        acc_weights = np.cumsum(weights)
        start = 0
        n = len(acc_weights)
        while start < n:
            # find the index of target_weight in the accumulated weights
            stop = int(np.searchsorted(acc_weights, target_weight, side="right"))
            # yield partition
            yield WeightedRange(values[start:stop], weights[start:stop])
            # update target_weight and start index
            target_weight = acc_weights[stop - 1] + max_weight
            start = stop



"""Functionality for loading data from TileDB arrays to the Tensorflow Data API."""

from typing import Sequence, Union

import sparse
import tensorflow as tf

from ._tensor_schema import DenseTensorSchema, SparseTensorSchema, TensorSchema
from .types import ArrayParams


def TensorflowTileDBDataset(
    *array_params: ArrayParams,
    batch_size: int,
    shuffle_buffer_size: int = 0,
    prefetch: int = tf.data.AUTOTUNE,
    num_workers: int = 0,
) -> tf.data.Dataset:
    """Return a tf.data.Dataset for loading data from TileDB arrays.

    :param array_params: One or more `ArrayParams` instances, one per TileDB array.
    :param batch_size: Size of each batch.
    :param shuffle_buffer_size: Number of elements from which this dataset will sample.
    :param prefetch: Maximum number of batches that will be buffered when prefetching.
        By default, the buffer size is dynamically tuned.
    :param num_workers: If greater than zero, create a threadpool of `num_workers` threads
        used to fetch inputs asynchronously and in parallel. Note: when `num_workers` > 1
        yielded batches may be shuffled even if `shuffle_buffer_size` is zero.
    """
    schemas = tuple(map(_get_tensor_schema, array_params))
    key_range = schemas[0].key_range
    if not all(key_range.equal_values(schema.key_range) for schema in schemas[1:]):
        raise ValueError(f"All arrays must have the same key range: {key_range}")

    max_weights = tuple(schema.max_partition_weight for schema in schemas)
    key_subranges = tuple(key_range.partition_by_count(num_workers or 1))

    def key_range_dataset(key_range_idx: int) -> tf.data.Dataset:
        datasets = tuple(
            tf.data.Dataset.from_generator(
                lambda i, schema=schema, max_weight=max_weight: schema.iter_tensors(
                    key_subranges[i].partition_by_weight(max_weight)
                ),
                args=(key_range_idx,),
                output_signature=_get_tensor_specs(schema),
            ).unbatch()
            for schema, max_weight in zip(schemas, max_weights)
        )
        return tf.data.Dataset.zip(datasets) if len(datasets) > 1 else datasets[0]

    if num_workers:
        dataset = tf.data.Dataset.from_tensor_slices(range(len(key_subranges)))
        dataset = dataset.interleave(
            key_range_dataset, num_parallel_calls=num_workers, deterministic=False
        )
    else:
        dataset = key_range_dataset(0)

    if shuffle_buffer_size > 0:
        dataset = dataset.shuffle(shuffle_buffer_size)
    return dataset.batch(batch_size).prefetch(prefetch)


TensorSpec = Union[tf.TensorSpec, tf.SparseTensorSpec]


def _get_tensor_specs(schema: TensorSchema) -> Union[TensorSpec, Sequence[TensorSpec]]:
    cls = tf.SparseTensorSpec if schema.sparse else tf.TensorSpec
    shape = (None, *schema.shape[1:])
    specs = tuple(cls(shape=shape, dtype=dtype) for dtype in schema.field_dtypes)
    return specs if len(specs) > 1 else specs[0]


def _get_tensor_schema(array_params: ArrayParams) -> TensorSchema:
    if not array_params.array.schema.sparse:
        return DenseTensorSchema.from_array_params(array_params)
    else:
        return SparseTensorSchema.from_array_params(array_params, _coo_to_sparse_tensor)


def _coo_to_sparse_tensor(coo: sparse.COO) -> tf.SparseTensor:
    return tf.SparseTensor(coo.coords.T, coo.data, coo.shape)



"""Functionality for loading data from TileDB arrays to the PyTorch Dataloader API."""

import itertools
import random
from operator import methodcaller
from typing import Callable, Iterable, Iterator, Sequence, Tuple, TypeVar, Union

import numpy as np
import scipy.sparse
import sparse
import torch
from torch.utils.data import DataLoader, IterableDataset, get_worker_info

try:
    # torch>=1.10
    sparse_csr_tensor = torch.sparse_csr_tensor
except AttributeError:
    # torch=1.9
    sparse_csr_tensor = torch._sparse_csr_tensor

import tiledb

from ._tensor_schema import DenseTensorSchema, SparseTensorSchema, TensorSchema
from .types import ArrayParams

Tensor = Union[np.ndarray, sparse.COO, scipy.sparse.csr_matrix]
TensorSequence = Union[
    Sequence[np.ndarray], Sequence[sparse.COO], Sequence[scipy.sparse.csr_matrix]
]
TensorOrSequence = Union[Tensor, TensorSequence]
OneOrMoreTensorsOrSequences = Union[TensorOrSequence, Tuple[TensorOrSequence, ...]]


def PyTorchTileDBDataLoader(
    *array_params: ArrayParams,
    batch_size: int,
    shuffle_buffer_size: int = 0,
    prefetch: int = 2,
    num_workers: int = 0,
    csr: bool = True,
) -> DataLoader:
    """Return a DataLoader for loading data from TileDB arrays.

    :param array_params: One or more `ArrayParams` instances, one per TileDB array.
    :param batch_size: Size of each batch.
    :param shuffle_buffer_size: Number of elements from which this dataset will sample.
    :param prefetch: Number of samples loaded in advance by each worker. Not applicable
        (and should not be given) when `num_workers` is 0.
    :param num_workers: how many subprocesses to use for data loading. 0 means that the
        data will be loaded in the main process. Note: when `num_workers` > 1
        yielded batches may be shuffled even if `shuffle_buffer_size` is zero.
    :param csr: For sparse 2D arrays, whether to return CSR tensors instead of COO.
    """
    schemas = tuple(map(_get_tensor_schema, array_params))
    collators = tuple(
        _get_tensor_collator(params.array, csr, len(schema.fields))
        for params, schema in zip(array_params, schemas)
    )
    collate_fn = _CompositeCollator(*collators) if len(collators) > 1 else collators[0]

    return DataLoader(
        dataset=_PyTorchTileDBDataset(schemas, shuffle_buffer_size=shuffle_buffer_size),
        batch_size=batch_size,
        prefetch_factor=prefetch,
        num_workers=num_workers,
        worker_init_fn=_worker_init,
        collate_fn=collate_fn,
    )


class _PyTorchTileDBDataset(IterableDataset[OneOrMoreTensorsOrSequences]):
    def __init__(self, schemas: Sequence[TensorSchema], shuffle_buffer_size: int = 0):
        super().__init__()
        key_range = schemas[0].key_range
        if not all(key_range.equal_values(schema.key_range) for schema in schemas[1:]):
            raise ValueError(f"All arrays must have the same key range: {key_range}")
        self.schemas = schemas
        self.key_range = key_range
        self._shuffle_buffer_size = shuffle_buffer_size

    def __iter__(self) -> Iterator[OneOrMoreTensorsOrSequences]:
        rows: Iterator[OneOrMoreTensorsOrSequences]
        it_rows = tuple(map(self._iter_rows, self.schemas))
        rows = zip(*it_rows) if len(it_rows) > 1 else it_rows[0]
        if self._shuffle_buffer_size > 0:
            rows = _iter_shuffled(rows, self._shuffle_buffer_size)
        return rows

    def _iter_rows(self, schema: TensorSchema) -> Iterator[TensorOrSequence]:
        max_weight = schema.max_partition_weight
        key_subranges = self.key_range.partition_by_weight(max_weight)
        batches: Iterable[TensorOrSequence] = schema.iter_tensors(key_subranges)
        if len(schema.fields) == 1:
            return (tensor for batch in batches for tensor in batch)
        else:
            return (tensors for batch in batches for tensors in zip(*batch))


def _worker_init(worker_id: int) -> None:
    worker_info = get_worker_info()
    dataset = worker_info.dataset
    if any(schema.sparse for schema in dataset.schemas):
        raise NotImplementedError("https://github.com/pytorch/pytorch/issues/20248")
    key_ranges = list(dataset.key_range.partition_by_count(worker_info.num_workers))
    dataset.key_range = key_ranges[worker_id]


def _get_tensor_schema(array_params: ArrayParams) -> TensorSchema:
    if not array_params.array.schema.sparse:
        return DenseTensorSchema.from_array_params(array_params)
    elif array_params.array.ndim == 2:
        return SparseTensorSchema.from_array_params(array_params, methodcaller("tocsr"))
    else:
        return SparseTensorSchema.from_array_params(array_params)


_SingleCollator = Callable[[TensorSequence], torch.Tensor]


class _CompositeCollator:
    """
    A callable for collating "rows" of data by a separate collator for each "column".
    Returns the collated columns collected into a tuple.
    """

    def __init__(self, *collators: _SingleCollator):
        self._collators = collators

    def __call__(self, rows: Sequence[TensorSequence]) -> Sequence[torch.Tensor]:
        columns = tuple(zip(*rows))
        collators = self._collators
        assert len(columns) == len(collators)
        return tuple(collator(column) for collator, column in zip(collators, columns))


def _ndarray_collate(arrays: Sequence[np.ndarray]) -> torch.Tensor:
    """Collate multiple Numpy arrays to a torch.Tensor with strided layout."""
    # Specialized version of default_collate for collating Numpy arrays
    # Faster than `torch.as_tensor(arrays)` (https://github.com/pytorch/pytorch/pull/51731)
    # and `torch.stack([torch.as_tensor(array) for array in arrays]])`
    return torch.from_numpy(np.stack(arrays))


def _sparse_coo_collate(arrays: Sequence[sparse.COO]) -> torch.Tensor:
    """Collate multiple sparse.COO arrays to a torch.Tensor with sparse_coo layout."""
    stacked = sparse.stack(arrays)
    return torch.sparse_coo_tensor(stacked.coords, stacked.data, stacked.shape)


def _csr_to_coo_collate(arrays: Sequence[scipy.sparse.csr_matrix]) -> torch.Tensor:
    """Collate multiple Scipy CSR matrices to a torch.Tensor with sparse_coo layout."""
    stacked = scipy.sparse.vstack(arrays).tocoo()
    coords = np.stack((stacked.row, stacked.col))
    return torch.sparse_coo_tensor(coords, stacked.data, stacked.shape)


def _csr_collate(arrays: Sequence[scipy.sparse.csr_matrix]) -> torch.Tensor:
    """Collate multiple Scipy CSR matrices to a torch.Tensor with sparse_csr layout."""
    stacked = scipy.sparse.vstack(arrays)
    return sparse_csr_tensor(
        torch.from_numpy(stacked.indptr),
        torch.from_numpy(stacked.indices),
        stacked.data,
        stacked.shape,
    )


def _get_tensor_collator(
    array: tiledb.Array, csr: bool, num_fields: int
) -> Union[_SingleCollator, _CompositeCollator]:
    if not array.schema.sparse:
        collator = _ndarray_collate
    elif array.ndim != 2:
        collator = _sparse_coo_collate
    elif csr:
        collator = _csr_collate
    else:
        collator = _csr_to_coo_collate

    if num_fields == 1:
        return collator
    else:
        return _CompositeCollator(*itertools.repeat(collator, num_fields))


_T = TypeVar("_T")


def _iter_shuffled(iterable: Iterable[_T], buffer_size: int) -> Iterator[_T]:
    """
    Shuffle the given iterable with a buffer.

    The buffer with `buffer_size` is filled with elements from the iterable first.
    Then, each item will be yielded from the buffer by reservoir sampling via iterator.

    """
    iterator = iter(iterable)
    buffer = list(itertools.islice(iterator, buffer_size))
    randrange = random.randrange
    for x in iterator:
        idx = randrange(0, buffer_size)
        yield buffer[idx]
        buffer[idx] = x
    random.shuffle(buffer)
    while buffer:
        yield buffer.pop()



import os
from typing import Any

import numpy as np
import tiledb.cloud

# Your TileDB username and password, exported as environmental variables
TILEDB_USER_NAME = os.environ.get("TILEDB_USER_NAME")
TILEDB_PASSWD = os.environ.get("TILEDB_PASSWD")

# Your TileDB namespace
TILEDB_NAMESPACE = "your_tiledb_namespace"

# Your S3 bucket
S3_BUCKET = "your_s3_bucket"

IMAGES_URI = "tiledb://{}/s3://{}/mnist_images".format(TILEDB_NAMESPACE, S3_BUCKET)
LABELS_URI = "tiledb://{}/s3://{}/mnist_labels".format(TILEDB_NAMESPACE, S3_BUCKET)


# Let's define an ingestion function
def ingest_in_tiledb(data: np.array, batch_size: int, uri: str) -> None:
    import tiledb

    # Equal number of dimensions with the numpy array.
    dims = [
        tiledb.Dim(
            name="dim_" + str(dim),
            domain=(0, data.shape[dim] - 1),
            tile=data.shape[dim] if dim > 0 else batch_size,
            dtype=np.int32,
        )
        for dim in range(data.ndim)
    ]

    # TileDB schema
    schema = tiledb.ArraySchema(
        domain=tiledb.Domain(*dims),
        sparse=False,
        attrs=[tiledb.Attr(name="features", dtype=data.dtype)],
    )
    # Create array
    tiledb.Array.create(uri, schema)

    # Ingest
    with tiledb.open(uri, "w") as tiledb_array:
        tiledb_array[:] = {"features": data}


def mnist_ingest(ingestion_func: Any) -> None:
    import torchvision

    train_data = torchvision.datasets.MNIST("./data", train=True, download=True)

    images = train_data.data.numpy() / 255.0
    labels = train_data.targets.numpy()

    # Ingest images
    ingestion_func(data=images, batch_size=64, uri=IMAGES_URI)

    # Ingest labels
    ingestion_func(data=labels, batch_size=64, uri=LABELS_URI)


tiledb.cloud.login(username=TILEDB_USER_NAME, password=TILEDB_PASSWD)

tiledb.cloud.udf.exec(mnist_ingest, ingestion_func=ingest_in_tiledb)

print(tiledb.cloud.last_udf_task().logs)



"""Functionality for saving and loading Tensorflow Keras models as TileDB arrays"""

import io
import json
import logging
import os
import pickle
from operator import attrgetter
from typing import Any, List, Mapping, Optional, Tuple

import numpy as np
import tensorflow as tf

import tiledb

from ._base import Meta, TileDBModel, Timestamp, current_milli_time
from ._cloud_utils import update_file_properties
from ._tensorboard import load_tensorboard, save_tensorboard

try:
    import keras

    if keras.Model is not tf.keras.Model:
        raise ImportError
    tf_keras_is_keras = True
except ImportError:
    import tensorflow.python.keras as keras

    tf_keras_is_keras = False

SharedObjectLoadingScope = keras.utils.generic_utils.SharedObjectLoadingScope
FunctionalOrSequential = (keras.models.Functional, keras.models.Sequential)
TFOptimizer = keras.optimizer_v1.TFOptimizer
get_json_type = keras.saving.saved_model.json_utils.get_json_type
preprocess_weights_for_loading = keras.saving.hdf5_format.preprocess_weights_for_loading
saving_utils = keras.saving.saving_utils


class TensorflowKerasTileDBModel(TileDBModel[tf.keras.Model]):
    """
    Class that implements all functionality needed to save Tensorflow models as
    TileDB arrays and load Tensorflow models from TileDB arrays.
    """

    Framework = "TENSORFLOW KERAS"
    FrameworkVersion = tf.__version__

    def save(
        self,
        *,
        update: bool = False,
        meta: Optional[Meta] = None,
        include_optimizer: bool = False,
        include_callbacks: Optional[tf.keras.callbacks.CallbackList] = None,
    ) -> None:
        """
        Save a Tensorflow model as a TileDB array.

        :param update: Whether we should update any existing TileDB array model at the
            target location.
        :param meta: Extra metadata to save in a TileDB array.
        :param include_optimizer: Whether to save the optimizer or not.
        :param include_callbacks: Callbacks list to store their data in array's metadata
        """
        if self.model is None:
            raise RuntimeError("Model is not initialized")

        # Used in this format only when model is Functional or Sequential
        model_weights = pickle.dumps(self.model.get_weights(), protocol=4)

        # Serialize model optimizer
        optimizer_weights = self._serialize_optimizer_weights(
            include_optimizer=include_optimizer
        )

        if include_callbacks:
            for cb in include_callbacks:
                if isinstance(cb, tf.keras.callbacks.TensorBoard):
                    cb_meta = save_tensorboard(os.path.join(cb.log_dir, "train"))
                    meta = {**meta, **cb_meta} if meta else cb_meta

        # Create TileDB model array
        if not update:
            self._create_array()

        self._write_array(
            include_optimizer=include_optimizer,
            serialized_weights=model_weights,
            serialized_optimizer_weights=optimizer_weights,
            meta=meta,
        )

    def load(
        self,
        *,
        timestamp: Optional[Timestamp] = None,
        compile_model: bool = False,
        custom_objects: Optional[Mapping[str, Any]] = None,
        input_shape: Optional[Tuple[int, ...]] = None,
    ) -> tf.keras.Model:
        """
        Load a Tensorflow model from a TileDB array.

        :param timestamp: Range of timestamps to load fragments of the array which live
            in the specified time range.
        :param compile_model: Whether to compile the model after loading or not.
        :param custom_objects: Mapping of names to custom classes or functions to be
            considered during deserialization.
        :param input_shape: The shape that the custom model expects as input
        :return: Tensorflow model.
        """
        # TODO: Change timestamp when issue in core is resolved

        with tiledb.open(self.uri, ctx=self.ctx, timestamp=timestamp) as model_array:
            model_array_results = model_array[:]
            model_config = json.loads(model_array.meta["model_config"])
            model_class = model_config["class_name"]

            if model_class not in ("Functional", "Sequential"):
                with SharedObjectLoadingScope():
                    with tf.keras.utils.CustomObjectScope(custom_objects or {}):
                        if hasattr(model_config, "decode"):
                            model_config = model_config.decode("utf-8")
                        model = tf.keras.models.model_from_config(
                            model_config, custom_objects=custom_objects
                        )
                        if not model.built:
                            model.build(input_shape)

                        # Load weights for layers
                        self._load_custom_subclassed_model(model, model_array)
            else:
                cls = (
                    tf.keras.Sequential
                    if model_class == "Sequential"
                    else tf.keras.Model
                )
                model = cls.from_config(model_config["config"])
                model_weights = pickle.loads(
                    model_array_results["model_weights"].item(0)
                )
                model.set_weights(model_weights)

            if compile_model:
                optimizer_weights = pickle.loads(
                    model_array_results["optimizer_weights"].item(0)
                )
                training_config = json.loads(model_array.meta["training_config"])

                # Compile model.
                model.compile(
                    **saving_utils.compile_args_from_training_config(
                        training_config, custom_objects
                    )
                )
                saving_utils.try_build_compiled_arguments(model)

                # Set optimizer weights.
                if optimizer_weights:
                    try:
                        model.optimizer._create_all_weights(model.trainable_variables)
                    except (NotImplementedError, AttributeError):
                        logging.warning(
                            "Error when creating the weights of optimizer {}, making it "
                            "impossible to restore the saved optimizer state. As a result, "
                            "your model is starting with a freshly initialized optimizer."
                        )

                    try:
                        model.optimizer.set_weights(optimizer_weights)
                    except ValueError:
                        logging.warning(
                            "Error in loading the saved optimizer "
                            "state. As a result, your model is "
                            "starting with a freshly initialized "
                            "optimizer."
                        )
            return model

    def load_tensorboard(
        self,
        target_dir: Optional[str] = None,
        timestamp: Optional[Timestamp] = None,
    ) -> None:
        return load_tensorboard(self.uri, self.ctx, target_dir, timestamp)

    def preview(self) -> str:
        """Create a string representation of the model."""
        if self.model:
            s = io.StringIO()
            self.model.summary(print_fn=lambda x: s.write(x + "\n"))
            model_summary = s.getvalue()
            return model_summary
        else:
            return ""

    def _create_array(self) -> None:
        """Create a TileDB array for a Tensorflow model"""
        assert self.model
        dom = tiledb.Domain(
            tiledb.Dim(
                name="model", domain=(1, 1), tile=1, dtype=np.int32, ctx=self.ctx
            )
            if isinstance(self.model, FunctionalOrSequential)
            else tiledb.Dim(
                name="model",
                domain=(1, len(self.model.layers)),
                tile=1,
                dtype=np.int32,
                ctx=self.ctx,
            ),
        )
        if isinstance(self.model, FunctionalOrSequential):
            attrs = [
                tiledb.Attr(
                    name="model_weights",
                    dtype=bytes,
                    var=True,
                    filters=tiledb.FilterList([tiledb.ZstdFilter()]),
                    ctx=self.ctx,
                ),
                tiledb.Attr(
                    name="optimizer_weights",
                    dtype=bytes,
                    var=True,
                    filters=tiledb.FilterList([tiledb.ZstdFilter()]),
                    ctx=self.ctx,
                ),
            ]
        else:
            attrs = [
                # String names of weights of each layer of the model
                tiledb.Attr(
                    name="weight_names",
                    dtype=bytes,
                    var=True,
                    filters=tiledb.FilterList([tiledb.ZstdFilter()]),
                    ctx=self.ctx,
                ),
                # The values of weights of each layer of the model
                tiledb.Attr(
                    name="weight_values",
                    dtype=bytes,
                    var=True,
                    filters=tiledb.FilterList([tiledb.ZstdFilter()]),
                    ctx=self.ctx,
                ),
                # Layer names TF format of the saved/loaded model
                tiledb.Attr(
                    name="layer_name",
                    dtype=str,
                    var=True,
                    filters=tiledb.FilterList([tiledb.ZstdFilter()]),
                    ctx=self.ctx,
                ),
                # The weight values of the optimizer in case the model is saved compiled
                tiledb.Attr(
                    name="optimizer_weights",
                    dtype=bytes,
                    var=True,
                    filters=tiledb.FilterList([tiledb.ZstdFilter()]),
                    ctx=self.ctx,
                ),
            ]

        schema = tiledb.ArraySchema(
            domain=dom,
            sparse=False,
            attrs=attrs,
            ctx=self.ctx,
        )

        tiledb.Array.create(self.uri, schema, ctx=self.ctx)

        # In case we are on TileDB-Cloud we have to update model array's file properties
        if self.namespace:
            update_file_properties(self.uri, self._file_properties)

    def _write_array(
        self,
        include_optimizer: bool,
        serialized_weights: bytes,
        serialized_optimizer_weights: bytes,
        meta: Optional[Meta],
    ) -> None:
        """Write Tensorflow model to a TileDB array."""
        assert self.model
        # TODO: Change timestamp when issue in core is resolved
        with tiledb.open(
            self.uri, "w", timestamp=current_milli_time(), ctx=self.ctx
        ) as tf_model_tiledb:
            if isinstance(self.model, FunctionalOrSequential):
                tf_model_tiledb[:] = {
                    "model_weights": np.array([serialized_weights]),
                    "optimizer_weights": np.array([serialized_optimizer_weights]),
                }
            else:
                # Insert weights and optimizer
                layer_names = []
                weight_names = []
                weight_values = []
                for layer in sorted(self.model.layers, key=attrgetter("name")):
                    weights = layer.trainable_weights + layer.non_trainable_weights
                    weight_values.append(
                        pickle.dumps(tf.keras.backend.batch_get_value(weights))
                    )
                    weight_names.append(
                        pickle.dumps([w.name.encode("utf8") for w in weights])
                    )
                    layer_names.append(layer.name)

                tf_model_tiledb[:] = {
                    "layer_name": np.array(layer_names),
                    "weight_values": np.array(weight_values),
                    "weight_names": np.array(weight_names),
                    # TODO (TeamML) Fix array scheme to avoid optimizer_weight repetition. Nullable
                    "optimizer_weights": np.array(
                        [
                            serialized_optimizer_weights
                            for _ in range(len(self.model.layers))
                        ]
                    ),
                }

            # Insert all model metadata
            model_metadata = saving_utils.model_metadata(self.model, include_optimizer)
            for key, value in model_metadata.items():
                tf_model_tiledb.meta[key] = json.dumps(
                    value, default=get_json_type
                ).encode("utf8")

            self.update_model_metadata(array=tf_model_tiledb, meta=meta)

    def _serialize_optimizer_weights(self, include_optimizer: bool = True) -> bytes:
        """Serialize optimizer weights"""
        assert self.model
        optimizer = self.model.optimizer
        if include_optimizer and optimizer and not isinstance(optimizer, TFOptimizer):
            optimizer_weights = tf.keras.backend.batch_get_value(optimizer.weights)
            return pickle.dumps(optimizer_weights, protocol=4)
        else:
            return b""

    def _load_custom_subclassed_model(
        self, model: tf.keras.Model, model_array: tiledb.Array
    ) -> None:
        if "keras_version" in model_array.meta:
            original_keras_version = model_array.meta["keras_version"]
            if hasattr(original_keras_version, "decode"):
                original_keras_version = original_keras_version.decode("utf8")
        else:
            original_keras_version = "1"
        if "backend" in model_array.meta:
            original_backend = model_array.meta["backend"]
            if hasattr(original_backend, "decode"):
                original_backend = original_backend.decode("utf8")
        else:
            original_backend = None

        # Load weights for layers
        self._load_weights_from_tiledb(
            model_array[:], model, original_keras_version, original_backend
        )

    def _load_weights_from_tiledb(
        self,
        model_array_results: Mapping[str, Any],
        model: tf.keras.Model,
        original_keras_version: Optional[str],
        original_backend: Optional[str],
    ) -> None:
        num_layers = 0
        for layer in model.layers:
            weights = layer.trainable_weights + layer.non_trainable_weights
            if weights:
                num_layers += 1

        read_layer_names = []
        for k, name in enumerate(model_array_results["layer_name"]):
            layer_weight_names = pickle.loads(
                model_array_results["weight_names"].item(k)
            )
            if layer_weight_names:
                read_layer_names.append(name)

        if len(read_layer_names) != num_layers:
            raise ValueError(
                f"You are trying to load a weight file with {len(read_layer_names)} "
                f"layers into a model with {num_layers} layers"
            )

        var_value_tuples: List[Tuple[tf.Variable, np.ndarray]] = []
        for k, layer in enumerate(model.layers):
            weight_vars = layer.trainable_weights + layer.non_trainable_weights
            read_weight_values = pickle.loads(
                model_array_results["weight_values"].item(k)
            )
            read_weight_values = preprocess_weights_for_loading(
                layer, read_weight_values, original_keras_version, original_backend
            )
            if len(read_weight_values) != len(weight_vars):
                raise ValueError(
                    f'Layer #{k}  (named "{layer.name}" in the current model) was found '
                    f"to correspond to layer {layer} in the save file. However the new "
                    f"layer {layer.name} expects {len(weight_vars)} weights, "
                    f"but the saved weights have {len(read_weight_values)} elements"
                )
            var_value_tuples.extend(zip(weight_vars, read_weight_values))
        tf.keras.backend.batch_set_value(var_value_tuples)



from dataclasses import dataclass, field
from typing import Any, Mapping, Sequence, TypeVar, Union

import tiledb

Tensor = TypeVar("Tensor")


@dataclass(frozen=True)
class ArrayParams:
    array: tiledb.Array
    key_dim: Union[int, str] = 0
    fields: Sequence[str] = ()
    _tensor_schema_kwargs: Mapping[str, Any] = field(init=False, repr=False)

    def __post_init__(self) -> None:
        all_attrs = [self.array.attr(i).name for i in range(self.array.nattr)]
        all_dims = [self.array.dim(i).name for i in range(self.array.ndim)]
        dims = []

        if self.fields:
            attrs = []
            for f in self.fields:
                if f in all_attrs:
                    attrs.append(f)
                elif f in all_dims:
                    dims.append(f)
                else:
                    raise ValueError(f"Unknown attribute or dimension '{field}'")
            final_fields = self.fields
        else:
            final_fields = attrs = all_attrs

        ned = list(self.array.nonempty_domain())
        key_dim_index = (
            self.key_dim
            if isinstance(self.key_dim, int)
            else all_dims.index(self.key_dim)
        )
        if key_dim_index > 0:
            # Swap key dimension to first position
            all_dims[0], all_dims[key_dim_index] = all_dims[key_dim_index], all_dims[0]
            ned[0], ned[key_dim_index] = ned[key_dim_index], ned[0]

        object.__setattr__(
            self,
            "_tensor_schema_kwargs",
            dict(
                array=self.array,
                fields=tuple(final_fields),
                key_dim_index=key_dim_index,
                ned=tuple(ned),
                all_dims=tuple(all_dims),
                query_kwargs={"attrs": tuple(attrs), "dims": tuple(dims)},
            ),
        )



import itertools as it
import os
import uuid
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Any, Sequence

import numpy as np
import pytest
import scipy.sparse
import sparse
import tensorflow as tf
import torch

import tiledb

NUM_ROWS = 107


def parametrize_for_dataset(
    *,
    x_shape=((NUM_ROWS, 10), (NUM_ROWS, 10, 3)),
    y_shape=((NUM_ROWS, 5), (NUM_ROWS, 5, 2)),
    x_sparse=(True, False),
    y_sparse=(True, False),
    key_dim_dtype=(np.dtype(np.int32), np.dtype("datetime64[D]"), np.dtype(np.bytes_)),
    x_key_dim=(0, 1),
    y_key_dim=(0, 1),
    num_fields=(0, 1, 2),
    batch_size=(8,),
    shuffle_buffer_size=(16,),
    num_workers=(0, 2),
):
    def is_valid_combination(t):
        _, _, x_sparse_, y_sparse_, key_dim_dtype_, *_ = t
        return bool(
            np.issubdtype(key_dim_dtype_, np.integer) or (x_sparse_ and y_sparse_)
        )

    argnames = [
        "x_shape",
        "y_shape",
        "x_sparse",
        "y_sparse",
        "key_dim_dtype",
        "x_key_dim",
        "y_key_dim",
        "num_fields",
        "batch_size",
        "shuffle_buffer_size",
        "num_workers",
    ]
    argvalues = it.product(
        x_shape,
        y_shape,
        x_sparse,
        y_sparse,
        key_dim_dtype,
        x_key_dim,
        y_key_dim,
        num_fields,
        batch_size,
        shuffle_buffer_size,
        num_workers,
    )
    return pytest.mark.parametrize(argnames, filter(is_valid_combination, argvalues))


@contextmanager
def ingest_in_tiledb(tmpdir, shape, sparse, key_dim_dtype, key_dim, num_fields):
    """Context manager for ingesting data into TileDB."""
    array_uuid = str(uuid.uuid4())
    uri = os.path.join(tmpdir, array_uuid)
    data = original_data = _rand_array(shape, sparse)
    if key_dim > 0:
        data = np.moveaxis(data, 0, key_dim)
    data_idx = np.arange(data.size).reshape(data.shape)

    transforms = []
    for i in range(data.ndim):
        n = data.shape[i]
        dtype = key_dim_dtype
        if i != key_dim:
            dtype = np.dtype("int32")
            # set the domain to (-n/2, n/2) to test negative domain indexing
            min_value = -(n // 2)
        elif np.issubdtype(key_dim_dtype, np.integer):
            # set the domain to (-n/2, n/2) to test negative domain indexing
            min_value = -(n // 2)
        elif np.issubdtype(key_dim_dtype, np.datetime64):
            min_value = np.datetime64("2022-06-15")
        elif np.issubdtype(key_dim_dtype, np.bytes_):
            min_value = b"a"
        else:
            assert False, key_dim_dtype
        transforms.append(_IndexTransformer(f"dim_{i}", n, min_value, dtype))

    dims = [transform.dim for transform in transforms]
    attrs = [
        tiledb.Attr(name="data", dtype=np.float32),
        tiledb.Attr(name="idx", dtype=np.int16),
    ]
    schema = tiledb.ArraySchema(domain=tiledb.Domain(*dims), attrs=attrs, sparse=sparse)
    tiledb.Array.create(uri, schema)

    with tiledb.open(uri, "w") as tiledb_array:
        if sparse:
            nz_idxs = np.nonzero(data)
            dim_idxs = tuple(
                transform(idx) for transform, idx in zip(transforms, nz_idxs)
            )
            tiledb_array[dim_idxs] = {"data": data[nz_idxs], "idx": data_idx[nz_idxs]}
        else:
            tiledb_array[:] = {"data": data, "idx": data_idx}

    all_fields = [f.name for f in dims + attrs]
    # exclude the key dimension from the fields if it is not an integer
    if not np.issubdtype(key_dim_dtype, np.integer):
        del all_fields[key_dim]
    fields = np.random.choice(all_fields, size=num_fields, replace=False).tolist()

    with tiledb.open(uri) as array:
        yield {
            "data": original_data,
            "array": array,
            "key_dim": key_dim,
            "fields": fields,
        }


def _rand_array(shape: Sequence[int], sparse: bool = False) -> np.ndarray:
    """Create a random array of the given shape.

    :param shape: Shape of the array.
    :param sparse:
      - If false, all values will be in the (0, 1) range.
      - If true, only one value per row will be non-zero, the rest will be 0.
    """
    if not sparse:
        return np.random.random(shape)

    rows, cols = shape[0], np.prod(shape[1:])
    a = np.zeros((rows, cols))
    col_idxs = np.random.choice(cols, size=rows)
    a[np.arange(rows), col_idxs] = np.random.random(rows)
    return a.reshape(shape)


@dataclass(frozen=True)
class _IndexTransformer:
    name: str
    size: int
    min_value: Any
    dtype: np.dtype

    @property
    def dim(self):
        return tiledb.Dim(
            name=self.name,
            domain=(self.min_value, self(self.size - 1)),
            tile=np.random.randint(1, self.size + 1),
            dtype=self.dtype,
        )

    def __call__(self, idx):
        if isinstance(self.min_value, bytes):
            transformed_idx = _bytes_to_int(self.min_value) + idx
            if isinstance(transformed_idx, np.ndarray):
                int_to_bytes = np.vectorize(_int_to_bytes)
            else:
                int_to_bytes = _int_to_bytes
            return int_to_bytes(transformed_idx)
        else:
            return self.min_value + idx


def _bytes_to_int(data: bytes) -> int:
    s = 0
    for i, b in enumerate(reversed(data)):
        s += b * 256**i
    return s


def _int_to_bytes(n: int) -> bytes:
    s = bytearray()
    while n > 0:
        n, m = divmod(n, 256)
        s.append(m)
    s.reverse()
    return bytes(s)


def validate_tensor_generator(
    generator, num_fields, x_sparse, y_sparse, x_shape, y_shape, batch_size=None
):
    for x_tensors, y_tensors in generator:
        for x_tensor in x_tensors if num_fields != 1 else [x_tensors]:
            _validate_tensor(x_tensor, x_sparse, x_shape[1:], batch_size)
        for y_tensor in y_tensors if num_fields != 1 else [y_tensors]:
            _validate_tensor(y_tensor, y_sparse, y_shape[1:], batch_size)


def _validate_tensor(tensor, expected_sparse, expected_row_shape, batch_size=None):
    if batch_size is None and not isinstance(tensor, scipy.sparse.spmatrix):
        row_shape = tensor.shape
    else:
        num_rows, *row_shape = tensor.shape
        if batch_size is None:
            assert isinstance(tensor, scipy.sparse.spmatrix)
            assert num_rows == 1
        else:
            # num_rows may be less than batch_size
            assert num_rows <= batch_size, (num_rows, batch_size)
    assert tuple(row_shape) == expected_row_shape
    assert _is_sparse(tensor) == expected_sparse


def _is_sparse(tensor):
    if isinstance(tensor, torch.Tensor):
        return tensor.layout in (torch.sparse_coo, torch.sparse_csr)
    if isinstance(tensor, (scipy.sparse.spmatrix, sparse.SparseArray, tf.SparseTensor)):
        return True
    if isinstance(tensor, (np.ndarray, tf.Tensor)):
        return False
    assert False, f"Unknown tensor type: {type(tensor)}"



"""Functionality for saving and loading PytTorch models as TileDB arrays"""

import pickle
from typing import Any, Mapping, Optional

import numpy as np
import torch
from torch.optim import Optimizer
from torch.utils.tensorboard import SummaryWriter

import tiledb

from ._base import Meta, TileDBModel, Timestamp, current_milli_time
from ._cloud_utils import update_file_properties
from ._tensorboard import load_tensorboard, save_tensorboard


class PyTorchTileDBModel(TileDBModel[torch.nn.Module]):
    """
    Class that implements all functionality needed to save PyTorch models as
    TileDB arrays and load PyTorch models from TileDB arrays.
    """

    Framework = "PYTORCH"
    FrameworkVersion = torch.__version__

    def __init__(
        self,
        uri: str,
        namespace: Optional[str] = None,
        ctx: Optional[tiledb.Ctx] = None,
        model: Optional[torch.nn.Module] = None,
        optimizer: Optional[Optimizer] = None,
    ):
        super().__init__(uri, namespace, ctx, model)
        self.optimizer = optimizer

    def save(
        self,
        *,
        update: bool = False,
        meta: Optional[Meta] = None,
        model_info: Optional[Mapping[str, Any]] = None,
        summary_writer: Optional[SummaryWriter] = None,
    ) -> None:
        """
        Save a PyTorch model as a TileDB array.

        :param update: Whether we should update any existing TileDB array model at the
            target location.
        :param meta: Extra metadata to save in a TileDB array.
        :param model_info: Contains model info like loss, epoch etc, that could be needed
            to save a model's general checkpoint for inference and/or resuming training.
        :param summary_path: Contains summary writer's path for storing tensorboard metadata
                                in array's metadata
        """
        if self.model is None:
            raise RuntimeError("Model is not initialized")

        # Serialize model information
        serialized_model_info = (
            {key: pickle.dumps(value, protocol=4) for key, value in model_info.items()}
            if model_info
            else {}
        )

        serialized_model_dict = {
            "model_state_dict": pickle.dumps(self.model.state_dict(), protocol=4)
        }

        serialized_optimizer_dict = (
            {
                "optimizer_state_dict": pickle.dumps(
                    self.optimizer.state_dict(), protocol=4
                )
            }
            if self.optimizer
            else {}
        )

        # Summary writer
        if summary_writer:
            cb_meta = save_tensorboard(summary_writer.log_dir)
            meta = {**meta, **cb_meta} if meta else cb_meta

        # Create TileDB model array
        if not update:
            self._create_array(serialized_model_info)

        self._write_array(
            {
                **serialized_model_dict,
                **serialized_optimizer_dict,
                **serialized_model_info,
            },
            meta=meta,
        )

    # FIXME: This method should change to return the model, not the model_info dict
    def load(  # type: ignore
        self,
        *,
        timestamp: Optional[Timestamp] = None,
        model: torch.nn.Module,
        optimizer: Optimizer,
    ) -> Optional[Mapping[str, Any]]:
        """
        Load a PyTorch model from a TileDB array.

        :param timestamp: Range of timestamps to load fragments of the array which live
            in the specified time range.
        :param model: A defined PyTorch model.
        :param optimizer: A defined PyTorch optimizer.
        :return: A dictionary with attributes other than model or optimizer state_dict.
        """

        # TODO: Change timestamp when issue in core is resolved
        model_array = tiledb.open(self.uri, ctx=self.ctx, timestamp=timestamp)
        model_array_results = model_array[:]
        schema = model_array.schema

        model_state_dict = pickle.loads(model_array_results["model_state_dict"].item(0))
        optimizer_state_dict = pickle.loads(
            model_array_results["optimizer_state_dict"].item(0)
        )

        # Load model's state and optimizer dictionaries
        model.load_state_dict(model_state_dict)
        optimizer.load_state_dict(optimizer_state_dict)

        # Get the rest of the attributes
        out_dict = {}
        for idx in range(schema.nattr):
            attr_name = schema.attr(idx).name
            if (
                schema.attr(idx).name != "model_state_dict"
                and schema.attr(idx).name != "optimizer_state_dict"
            ):
                out_dict[attr_name] = pickle.loads(
                    model_array_results[attr_name].item(0)
                )
        return out_dict

    def load_tensorboard(
        self,
        target_dir: Optional[str] = None,
        timestamp: Optional[Timestamp] = None,
    ) -> None:
        return load_tensorboard(self.uri, self.ctx, target_dir, timestamp)

    def preview(self) -> str:
        """
        Create a string representation of the model.

        :return: str. A string representation of the models internal configuration.
        """
        return str(self.model) if self.model else ""

    def _create_array(self, serialized_model_info: Mapping[str, bytes]) -> None:
        """
        Create a TileDB array for a PyTorch model

        :param serialized_model_info: A mapping with pickled information of a PyTorch model.
        """
        dom = tiledb.Domain(
            tiledb.Dim(
                name="model", domain=(1, 1), tile=1, dtype=np.int32, ctx=self.ctx
            ),
        )

        attrs = []

        # Keep model's state dictionary
        attrs.append(
            tiledb.Attr(
                name="model_state_dict",
                dtype=bytes,
                var=True,
                filters=tiledb.FilterList([tiledb.ZstdFilter()]),
                ctx=self.ctx,
            ),
        )

        # If optimizer is provided we also keep optimizer's state dictionary
        if self.optimizer:
            attrs.append(
                tiledb.Attr(
                    name="optimizer_state_dict",
                    dtype=bytes,
                    var=True,
                    filters=tiledb.FilterList([tiledb.ZstdFilter()]),
                    ctx=self.ctx,
                ),
            )

        # Add extra attributes in case model information is provided by the user
        if serialized_model_info:
            for key in serialized_model_info:
                attrs.append(
                    tiledb.Attr(
                        name=key,
                        dtype=bytes,
                        var=True,
                        filters=tiledb.FilterList([tiledb.ZstdFilter()]),
                        ctx=self.ctx,
                    ),
                )

        schema = tiledb.ArraySchema(
            domain=dom,
            sparse=False,
            attrs=attrs,
            ctx=self.ctx,
        )

        tiledb.Array.create(self.uri, schema, ctx=self.ctx)

        # In case we are on TileDB-Cloud we have to update model array's file properties
        if self.namespace:
            update_file_properties(self.uri, self._file_properties)

    def _write_array(
        self, serialized_model_dict: Mapping[str, bytes], meta: Optional[Meta]
    ) -> None:
        """
        Write a PyTorch model to a TileDB array.

        :param serialized_model_dict: A mapping with pickled information (model state,
            optimizer state, extra model information) of a PyTorch model.
        :param meta: Extra metadata to save in a TileDB array.
        """

        # TODO: Change timestamp when issue in core is resolved
        with tiledb.open(
            self.uri, "w", timestamp=current_milli_time(), ctx=self.ctx
        ) as tf_model_tiledb:
            # Insertion in TileDB array
            tf_model_tiledb[:] = {
                key: np.array([value]) for key, value in serialized_model_dict.items()
            }
            self.update_model_metadata(array=tf_model_tiledb, meta=meta)



"""Tests for TileDB integration with Tensorflow Data API."""

import numpy as np
import pytest
import tensorflow as tf

from tiledb.ml.readers.tensorflow import TensorflowTileDBDataset
from tiledb.ml.readers.types import ArrayParams

from .utils import ingest_in_tiledb, parametrize_for_dataset, validate_tensor_generator


class TestTensorflowTileDBDataset:
    @parametrize_for_dataset()
    def test_dataset(
        self,
        tmpdir,
        x_shape,
        y_shape,
        x_sparse,
        y_sparse,
        key_dim_dtype,
        x_key_dim,
        y_key_dim,
        num_fields,
        batch_size,
        shuffle_buffer_size,
        num_workers,
    ):
        with ingest_in_tiledb(
            tmpdir, x_shape, x_sparse, key_dim_dtype, x_key_dim, num_fields
        ) as x_kwargs, ingest_in_tiledb(
            tmpdir, y_shape, y_sparse, key_dim_dtype, y_key_dim, num_fields
        ) as y_kwargs:
            dataset = TensorflowTileDBDataset(
                ArrayParams(x_kwargs["array"], x_kwargs["key_dim"], x_kwargs["fields"]),
                ArrayParams(y_kwargs["array"], y_kwargs["key_dim"], y_kwargs["fields"]),
                batch_size=batch_size,
                shuffle_buffer_size=shuffle_buffer_size,
                num_workers=num_workers,
            )
            assert isinstance(dataset, tf.data.Dataset)
            validate_tensor_generator(
                dataset, num_fields, x_sparse, y_sparse, x_shape, y_shape, batch_size
            )

    @parametrize_for_dataset(
        # Add one extra key on X
        x_shape=((108, 10), (108, 10, 3)),
        y_shape=((107, 5), (107, 5, 2)),
    )
    def test_unequal_num_keys(
        self,
        tmpdir,
        x_shape,
        y_shape,
        x_sparse,
        y_sparse,
        key_dim_dtype,
        x_key_dim,
        y_key_dim,
        num_fields,
        batch_size,
        shuffle_buffer_size,
        num_workers,
    ):
        with ingest_in_tiledb(
            tmpdir, x_shape, x_sparse, key_dim_dtype, x_key_dim, num_fields
        ) as x_kwargs, ingest_in_tiledb(
            tmpdir, y_shape, y_sparse, key_dim_dtype, y_key_dim, num_fields
        ) as y_kwargs:
            with pytest.raises(ValueError) as ex:
                TensorflowTileDBDataset(
                    ArrayParams(
                        x_kwargs["array"], x_kwargs["key_dim"], x_kwargs["fields"]
                    ),
                    ArrayParams(
                        y_kwargs["array"], y_kwargs["key_dim"], y_kwargs["fields"]
                    ),
                    batch_size=batch_size,
                    shuffle_buffer_size=shuffle_buffer_size,
                    num_workers=num_workers,
                )
            assert "All arrays must have the same key range" in str(ex.value)

    @parametrize_for_dataset(num_fields=[0], shuffle_buffer_size=[0], num_workers=[0])
    def test_dataset_order(
        self,
        tmpdir,
        x_shape,
        y_shape,
        x_sparse,
        y_sparse,
        key_dim_dtype,
        x_key_dim,
        y_key_dim,
        num_fields,
        batch_size,
        shuffle_buffer_size,
        num_workers,
    ):
        """Test we can read the data in the same order as written.

        The order is guaranteed only for sequential processing (num_workers=0) and
        no shuffling (shuffle_buffer_size=0).
        """
        with ingest_in_tiledb(
            tmpdir, x_shape, x_sparse, key_dim_dtype, x_key_dim, num_fields
        ) as x_kwargs, ingest_in_tiledb(
            tmpdir, y_shape, y_sparse, key_dim_dtype, y_key_dim, num_fields
        ) as y_kwargs:
            dataset = TensorflowTileDBDataset(
                ArrayParams(x_kwargs["array"], x_kwargs["key_dim"], x_kwargs["fields"]),
                ArrayParams(y_kwargs["array"], y_kwargs["key_dim"], y_kwargs["fields"]),
                batch_size=batch_size,
                shuffle_buffer_size=shuffle_buffer_size,
                num_workers=num_workers,
            )
            # since num_fields is 0, fields are all the array attributes of each array
            # the first item of each batch corresponds to the first attribute (="data")
            x_data_batches, y_data_batches = [], []
            for x_tensors, y_tensors in dataset:
                x_data_batch = x_tensors[0]
                if x_sparse:
                    x_data_batch = tf.sparse.to_dense(x_data_batch)
                x_data_batches.append(x_data_batch)

                y_data_batch = y_tensors[0]
                if y_sparse:
                    y_data_batch = tf.sparse.to_dense(y_data_batch)
                y_data_batches.append(y_data_batch)

            np.testing.assert_array_almost_equal(
                np.concatenate(x_data_batches), x_kwargs["data"]
            )
            np.testing.assert_array_almost_equal(
                np.concatenate(y_data_batches), y_kwargs["data"]
            )



import os
from typing import List

import tiledb.cloud

# Your TileDB username and password, exported as environmental variables
TILEDB_USER_NAME = os.environ.get("TILEDB_USER_NAME")
TILEDB_PASSWD = os.environ.get("TILEDB_PASSWD")

# Your TileDB namespace
TILEDB_NAMESPACE = "your_tiledb_namespace"

# Your S3 bucket
S3_BUCKET = "your_s3_bucket"

IMAGES_URI = "tiledb://{}/s3://{}/mnist_images".format(TILEDB_NAMESPACE, S3_BUCKET)
LABELS_URI = "tiledb://{}/s3://{}/mnist_labels".format(TILEDB_NAMESPACE, S3_BUCKET)
MODEL_URI = "tiledb://{}/s3://{}/mnist_model".format(TILEDB_NAMESPACE, S3_BUCKET)

IO_BATCH_SIZE = 20000


def predict() -> List[int]:
    from typing import Any, Tuple

    import numpy as np
    import torch
    import torch.nn as nn
    import torch.optim as optim

    from tiledb.ml.models.pytorch import PyTorchTileDBModel
    from tiledb.ml.readers.pytorch import PyTorchTileDBDataLoader

    class Net(nn.Module):
        def __init__(self, shape: Tuple[int, int]):
            super().__init__()
            self.flatten = nn.Flatten()
            self.linear_relu_stack = nn.Sequential(
                nn.Linear(np.product(shape), 32),
                nn.ReLU(),
                nn.Linear(32, 32),
                nn.ReLU(),
                nn.Linear(32, 10),
                nn.ReLU(),
            )

        def forward(self, x: torch.Tensor) -> Any:
            x = self.flatten(x)
            logits = self.linear_relu_stack(x)
            return logits

    # Load our model from a TileDB array
    loaded_model = Net(shape=(28, 28))
    loaded_optimizer = optim.SGD(loaded_model.parameters(), lr=0.01, momentum=0.5)

    PyTorchTileDBModel(uri=MODEL_URI).load(
        model=loaded_model, optimizer=loaded_optimizer
    )

    with tiledb.open(IMAGES_URI) as x, tiledb.open(LABELS_URI) as y:
        with torch.no_grad():
            data_loader = PyTorchTileDBDataLoader(x, y, batch_size=IO_BATCH_SIZE)
            inputs, labels = next(iter(data_loader))
            output = loaded_model(
                inputs[np.random.randint(0, 10) : np.random.randint(11, 50)].to(
                    torch.float
                )
            )

    return [np.argmax(pred) for pred in output.numpy()]


tiledb.cloud.login(username=TILEDB_USER_NAME, password=TILEDB_PASSWD)

predictions = tiledb.cloud.udf.exec(predict)

print(predictions)
